@@ -89,6 +89,8 @@ struct rfcomm_channel {
 
 static bdaddr_t adapter_addr;
 
+static uint8_t hal_mode = HAL_MODE_SOCKET_DEFAULT;
+
 static const uint8_t zero_uuid[16] = { 0 };
 
 /* Simple list of RFCOMM connected sockets */
@@ -313,10 +315,31 @@ static sdp_record_t *create_mas_record(uint8_t chan, const char *svc_name)
 {
 	sdp_list_t *seq;
 	sdp_profile_desc_t profile[1];
-	uint8_t minst = DEFAULT_MAS_INSTANCE;
-	uint8_t mtype = DEFAULT_MAS_MSG_TYPE;
+	uint8_t minst, mtype;
 	sdp_record_t *record;
 	uuid_t uuid;
+	int cnt, ret;
+
+	switch (hal_mode) {
+	case HAL_MODE_SOCKET_DYNAMIC_MAP:
+		/*
+		 * Service name for MAP is passed as XXYYname
+		 * XX - instance
+		 * YY - message type
+		 */
+		ret = sscanf(svc_name, "%02hhx%02hhx%n", &minst, &mtype, &cnt);
+		if (ret != 2 || cnt != 4)
+			return NULL;
+
+		svc_name += 4;
+		break;
+	case HAL_MODE_SOCKET_DEFAULT:
+		minst = DEFAULT_MAS_INSTANCE;
+		mtype = DEFAULT_MAS_MSG_TYPE;
+		break;
+	default:
+		return NULL;
+	}
 
 	sdp_uuid16_create(&uuid, MAP_MSE_SVCLASS_ID);
 
@@ -360,6 +383,7 @@ static sdp_record_t *create_app_record(uint8_t chan,
 	uuid_t uuid;
 
 	sdp_uuid128_create(&uuid, app_uuid);
+	sdp_uuid128_to_uuid(&uuid);
 
 	record = create_rfcomm_record(chan, &uuid, svc_name, false);
 	if (!record)
@@ -735,7 +759,8 @@ static int find_free_channel(void)
 
 static BtIOSecLevel get_sec_level(uint8_t flags)
 {
-	/* HAL_SOCK_FLAG_AUTH should require MITM but in our case setting
+	/*
+	 * HAL_SOCK_FLAG_AUTH should require MITM but in our case setting
 	 * security to BT_IO_SEC_HIGH would also require 16-digits PIN code
 	 * for pre-2.1 devices which is not what Android expects. For this
 	 * reason we ignore this flag to not break apps which use "secure"
@@ -1145,7 +1170,10 @@ void bt_socket_register(struct ipc *ipc, const bdaddr_t *addr, uint8_t mode)
 
 	DBG("");
 
-	/* make sure channels assigned for profiles are reserved and not used
+	hal_mode = mode;
+
+	/*
+	 * make sure channels assigned for profiles are reserved and not used
 	 * for app services
 	 */
 	for (i = 0; i < G_N_ELEMENTS(profiles); i++)
