@@ -66,12 +66,7 @@ struct bt_att {
 	bool writer_active;
 
 	struct queue *notify_list;	/* List of registered callbacks */
-	bool in_notify;
-	bool need_notify_cleanup;
-
 	struct queue *disconn_list;	/* List of disconnect handlers */
-	bool in_disconn;
-	bool need_disconn_cleanup;
 
 	bool in_req;			/* There's a pending incoming request */
 
@@ -216,7 +211,6 @@ static void cancel_att_send_op(struct att_send_op *op)
 struct att_notify {
 	unsigned int id;
 	uint16_t opcode;
-	bool removed;
 	bt_att_notify_func_t callback;
 	bt_att_destroy_func_t destroy;
 	void *user_data;
@@ -240,20 +234,6 @@ static bool match_notify_id(const void *a, const void *b)
 	return notify->id == id;
 }
 
-static bool match_notify_removed(const void *a, const void *b)
-{
-	const struct att_notify *notify = a;
-
-	return notify->removed;
-}
-
-static void mark_notify_removed(void *data, void *user_data)
-{
-	struct att_notify *notify = data;
-
-	notify->removed = true;
-}
-
 struct att_disconn {
 	unsigned int id;
 	bool removed;
@@ -280,20 +260,6 @@ static bool match_disconn_id(const void *a, const void *b)
 	return disconn->id == id;
 }
 
-static bool match_disconn_removed(const void *a, const void *b)
-{
-	const struct att_disconn *disconn = a;
-
-	return disconn->removed;
-}
-
-static void mark_disconn_removed(void *data, void *user_data)
-{
-	struct att_disconn *disconn = data;
-
-	disconn->removed = true;
-}
-
 static bool encode_pdu(struct att_send_op *op, const void *pdu,
 						uint16_t length, uint16_t mtu)
 {
@@ -536,36 +502,42 @@ static void wakeup_writer(struct bt_att *att)
 static void disconn_handler(void *data, void *user_data)
 {
 	struct att_disconn *disconn = data;
+	int err = PTR_TO_INT(user_data);
 
 	if (disconn->removed)
 		return;
 
 	if (disconn->callback)
-		disconn->callback(disconn->user_data);
+		disconn->callback(err, disconn->user_data);
 }
 
 static bool disconnect_cb(struct io *io, void *user_data)
 {
 	struct bt_att *att = user_data;
+	int err;
+	socklen_t len;
+
+	len = sizeof(err);
+
+	if (getsockopt(att->fd, SOL_SOCKET, SO_ERROR, &err, &len) < 0) {
+		util_debug(att->debug_callback, att->debug_data,
+					"Failed to obtain disconnect error: %s",
+					strerror(errno));
+		err = 0;
+	}
+
+	util_debug(att->debug_callback, att->debug_data,
+					"Physical link disconnected: %s",
+					strerror(err));
 
 	io_destroy(att->io);
 	att->io = NULL;
 
-	util_debug(att->debug_callback, att->debug_data,
-						"Physical link disconnected");
-
 	bt_att_cancel_all(att);
 
 	bt_att_ref(att);
-	att->in_disconn = true;
-	queue_foreach(att->disconn_list, disconn_handler, NULL);
-	att->in_disconn = false;
 
-	if (att->need_disconn_cleanup) {
-		queue_remove_all(att->disconn_list, match_disconn_removed, NULL,
-							destroy_att_disconn);
-		att->need_disconn_cleanup = false;
-	}
+	queue_foreach(att->disconn_list, disconn_handler, INT_TO_PTR(err));
 
 	bt_att_unregister_all(att);
 	bt_att_unref(att);
@@ -680,9 +652,6 @@ static void notify_handler(void *data, void *user_data)
 	struct att_notify *notify = data;
 	struct notify_data *not_data = user_data;
 
-	if (notify->removed)
-		return;
-
 	if (!opcode_match(notify->opcode, not_data->opcode))
 		return;
 
@@ -712,7 +681,6 @@ static void handle_notify(struct bt_att *att, uint8_t opcode, uint8_t *pdu,
 	struct notify_data data;
 
 	bt_att_ref(att);
-	att->in_notify = true;
 
 	memset(&data, 0, sizeof(data));
 	data.opcode = opcode;
@@ -724,22 +692,14 @@ static void handle_notify(struct bt_att *att, uint8_t opcode, uint8_t *pdu,
 
 	queue_foreach(att->notify_list, notify_handler, &data);
 
-	att->in_notify = false;
-
-	if (att->need_notify_cleanup) {
-		queue_remove_all(att->notify_list, match_notify_removed, NULL,
-							destroy_att_notify);
-		att->need_notify_cleanup = false;
-	}
-
-	bt_att_unref(att);
-
 	/*
 	 * If this was a request and no handler was registered for it, respond
 	 * with "Not Supported"
 	 */
 	if (!data.handler_found && get_op_type(opcode) == ATT_OP_TYPE_REQ)
 		respond_not_supported(att, opcode);
+
+	bt_att_unref(att);
 }
 
 static bool can_read_data(struct io *io, void *user_data)
@@ -762,6 +722,8 @@ static bool can_read_data(struct io *io, void *user_data)
 	pdu = att->buf;
 	opcode = pdu[0];
 
+	bt_att_ref(att);
+
 	/* Act on the received PDU based on the opcode type */
 	switch (get_op_type(opcode)) {
 	case ATT_OP_TYPE_RSP:
@@ -785,6 +747,7 @@ static bool can_read_data(struct io *io, void *user_data)
 					"Received request while another is "
 					"pending: 0x%02x", opcode);
 			io_shutdown(att->io);
+			bt_att_unref(att);
 
 			return false;
 		}
@@ -806,9 +769,38 @@ static bool can_read_data(struct io *io, void *user_data)
 		break;
 	}
 
+	bt_att_unref(att);
+
 	return true;
 }
 
+static void bt_att_free(struct bt_att *att)
+{
+	if (att->pending_req)
+		destroy_att_send_op(att->pending_req);
+
+	if (att->pending_ind)
+		destroy_att_send_op(att->pending_ind);
+
+	io_destroy(att->io);
+
+	queue_destroy(att->req_queue, NULL);
+	queue_destroy(att->ind_queue, NULL);
+	queue_destroy(att->write_queue, NULL);
+	queue_destroy(att->notify_list, NULL);
+	queue_destroy(att->disconn_list, NULL);
+
+	if (att->timeout_destroy)
+		att->timeout_destroy(att->timeout_data);
+
+	if (att->debug_destroy)
+		att->debug_destroy(att->debug_data);
+
+	free(att->buf);
+
+	free(att);
+}
+
 struct bt_att *bt_att_new(int fd)
 {
 	struct bt_att *att;
@@ -860,14 +852,7 @@ struct bt_att *bt_att_new(int fd)
 	return bt_att_ref(att);
 
 fail:
-	queue_destroy(att->req_queue, NULL);
-	queue_destroy(att->ind_queue, NULL);
-	queue_destroy(att->write_queue, NULL);
-	queue_destroy(att->notify_list, NULL);
-	queue_destroy(att->disconn_list, NULL);
-	io_destroy(att->io);
-	free(att->buf);
-	free(att);
+	bt_att_free(att);
 
 	return NULL;
 }
@@ -893,35 +878,7 @@ void bt_att_unref(struct bt_att *att)
 	bt_att_unregister_all(att);
 	bt_att_cancel_all(att);
 
-	if (att->pending_req)
-		destroy_att_send_op(att->pending_req);
-
-	if (att->pending_ind)
-		destroy_att_send_op(att->pending_ind);
-
-	io_destroy(att->io);
-	att->io = NULL;
-
-	queue_destroy(att->req_queue, NULL);
-	queue_destroy(att->ind_queue, NULL);
-	queue_destroy(att->write_queue, NULL);
-	queue_destroy(att->notify_list, NULL);
-	queue_destroy(att->disconn_list, NULL);
-	att->req_queue = NULL;
-	att->ind_queue = NULL;
-	att->write_queue = NULL;
-	att->notify_list = NULL;
-
-	if (att->timeout_destroy)
-		att->timeout_destroy(att->timeout_data);
-
-	if (att->debug_destroy)
-		att->debug_destroy(att->debug_data);
-
-	free(att->buf);
-	att->buf = NULL;
-
-	free(att);
+	bt_att_free(att);
 }
 
 bool bt_att_set_close_on_unref(struct bt_att *att, bool do_close)
@@ -1033,20 +990,12 @@ bool bt_att_unregister_disconnect(struct bt_att *att, unsigned int id)
 	if (!att || !id)
 		return false;
 
-	disconn = queue_find(att->disconn_list, match_disconn_id,
+	disconn = queue_remove_if(att->disconn_list, match_disconn_id,
 							UINT_TO_PTR(id));
 	if (!disconn)
 		return false;
 
-	if (!att->in_disconn) {
-		queue_remove(att->disconn_list, disconn);
-		destroy_att_disconn(disconn);
-		return true;
-	}
-
-	disconn->removed = true;
-	att->need_disconn_cleanup = true;
-
+	destroy_att_disconn(disconn);
 	return true;
 }
 
@@ -1258,20 +1207,12 @@ bool bt_att_unregister(struct bt_att *att, unsigned int id)
 	if (!att || !id)
 		return false;
 
-	notify = queue_find(att->notify_list, match_notify_id,
+	notify = queue_remove_if(att->notify_list, match_notify_id,
 							UINT_TO_PTR(id));
 	if (!notify)
 		return false;
 
-	if (!att->in_notify) {
-		queue_remove(att->notify_list, notify);
-		destroy_att_notify(notify);
-		return true;
-	}
-
-	notify->removed = true;
-	att->need_notify_cleanup = true;
-
+	destroy_att_notify(notify);
 	return true;
 }
 
@@ -1280,21 +1221,8 @@ bool bt_att_unregister_all(struct bt_att *att)
 	if (!att)
 		return false;
 
-	if (att->in_notify) {
-		queue_foreach(att->notify_list, mark_notify_removed, NULL);
-		att->need_notify_cleanup = true;
-	} else {
-		queue_remove_all(att->notify_list, NULL, NULL,
-							destroy_att_notify);
-	}
-
-	if (att->in_disconn) {
-		queue_foreach(att->disconn_list, mark_disconn_removed, NULL);
-		att->need_disconn_cleanup = true;
-	} else {
-		queue_remove_all(att->disconn_list, NULL, NULL,
-							destroy_att_disconn);
-	}
+	queue_remove_all(att->notify_list, NULL, NULL, destroy_att_notify);
+	queue_remove_all(att->disconn_list, NULL, NULL, destroy_att_disconn);
 
 	return true;
 }
