@@ -76,12 +76,6 @@ static void print_dev_list(int ctl, int flags)
 		di.dev_id = (dr+i)->dev_id;
 		if (ioctl(ctl, HCIGETDEVINFO, (void *) &di) < 0)
 			continue;
-		if (hci_test_bit(HCI_RAW, &di.flags) &&
-				!bacmp(&di.bdaddr, BDADDR_ANY)) {
-			int dd = hci_open_dev(di.dev_id);
-			hci_read_bd_addr(dd, &di.bdaddr, 1000);
-			hci_close_dev(dd);
-		}
 		print_dev_info(ctl, &di);
 	}
 }
@@ -630,6 +624,9 @@ static void cmd_features(int ctl, int hdev, char *opt)
 		exit(1);
 	}
 
+	if (max_page < 1 && (features[6] & LMP_SIMPLE_PAIR))
+		max_page = 1;
+
 	print_dev_hdr(&di);
 	printf("\tFeatures%s: 0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x "
 				"0x%2.2x 0x%2.2x 0x%2.2x 0x%2.2x\n",
@@ -856,6 +853,7 @@ static char *get_minor_device_name(int major, int minor)
 		}
 		if (strlen(cls_str) > 0)
 			return cls_str;
+		break;
 	}
 	case 6:	/* imaging */
 		if (minor & 4)
@@ -2016,13 +2014,6 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
-	if (hci_test_bit(HCI_RAW, &di.flags) &&
-			!bacmp(&di.bdaddr, BDADDR_ANY)) {
-		int dd = hci_open_dev(di.dev_id);
-		hci_read_bd_addr(dd, &di.bdaddr, 1000);
-		hci_close_dev(dd);
-	}
-
 	while (argc > 0) {
 		for (i = 0; command[i].cmd; i++) {
 			if (strncmp(command[i].cmd,
